import { ProtectReservedNames } from "./protect-types.js";
import { EventEmitter } from "node:events";
import { PROTECT_DOORBELL_TRIGGER_DURATION } from "./settings.js";
export class ProtectNvrEvents extends EventEmitter {
    api;
    hap;
    log;
    mqttPublishTelemetry;
    nvr;
    eventTimers;
    ufpApi;
    ufpDeviceState;
    platform;
    unsupportedDevices;
    eventsHandler;
    ufpUpdatesHandler;
    // Initialize an instance of our Protect events handler.
    constructor(nvr) {
        super();
        this.api = nvr.platform.api;
        this.eventTimers = {};
        this.hap = nvr.platform.api.hap;
        this.log = nvr.log;
        this.mqttPublishTelemetry = nvr.hasFeature("Nvr.Publish.Telemetry");
        this.nvr = nvr;
        this.ufpApi = nvr.ufpApi;
        this.ufpDeviceState = {};
        this.platform = nvr.platform;
        this.unsupportedDevices = {};
        this.eventsHandler = null;
        this.ufpUpdatesHandler = null;
        // If we've enabled telemetry from the controller inform the user.
        if (this.mqttPublishTelemetry) {
            this.log.info("Protect controller telemetry enabled.");
        }
        this.configureEvents();
    }
    // Thanks to https://stackoverflow.com/a/48218209 for the foundation for this one. Merge Protect JSON update payloads into the Protect configuration JSON for a device
    // while dealing with deep objects.
    //
    // @param {...object} objects - Objects to merge
    // @returns {object} - New object with merged key/values
    patchUfpConfigJson(...objects) {
        const isObject = (x) => (x && (typeof (x) === "object"));
        return objects.reduce((prev, obj) => {
            for (const key of Object.keys(obj)) {
                const pVal = prev[key];
                const oVal = obj[key];
                if (Array.isArray(pVal) && Array.isArray(oVal)) {
                    prev[key] = oVal;
                }
                else if (isObject(pVal) && isObject(oVal)) {
                    prev[key] = this.patchUfpConfigJson(pVal, oVal);
                }
                else {
                    prev[key] = oVal;
                }
            }
            return prev;
        }, {});
    }
    // Process Protect API update events.
    ufpUpdates(packet) {
        let protectDevice;
        switch (packet.header.modelKey) {
            case "nvr":
                this.nvr.ufp = this.patchUfpConfigJson(this.nvr.ufp, packet.payload);
                break;
            default:
                // Lookup the device.
                protectDevice = this.nvr.deviceLookup(packet.header.id);
                // Update our device state.
                if (protectDevice) {
                    protectDevice.ufp = this.patchUfpConfigJson(protectDevice.ufp, packet.payload);
                    // Sync names, if configured to do so.
                    if (packet.payload.name && protectDevice.hints.syncName) {
                        protectDevice.log.info("Name change detected. A restart of Homebridge may be needed in order to complete name synchronization with HomeKit.");
                        protectDevice.configureInfo();
                    }
                }
                break;
        }
        // Update the internal list we maintain.
        this.ufpDeviceState[packet.header.id] = Object.assign(this.ufpDeviceState[packet.header.id] ?? {}, packet.payload);
    }
    // Process device additions and removals from the Protect update events API.
    manageDevices(packet) {
        const payload = packet.payload;
        // We only want adoption-related events.
        if ((packet.header.modelKey !== "event") || ((payload.type !== "deviceAdopted") && (payload.type !== "deviceUnadopted"))) {
            return;
        }
        // Make sure we have the right information to process the event.
        if (!("deviceId" in payload.metadata) || !("text" in payload.metadata.deviceId)) {
            return;
        }
        // Lookup the device.
        const deviceId = payload.metadata.deviceId.text;
        const protectDevice = this.nvr.deviceLookup(deviceId);
        // We're adopting.
        if (payload.type === "deviceAdopted") {
            if (protectDevice) {
                return;
            }
            this.nvr.addHomeKitDevice(this.ufpDeviceState[deviceId]);
            return;
        }
        // We're unadopting.
        if (payload.type === "deviceUnadopted") {
            // If it's already gone, we're done.
            if (!protectDevice) {
                return;
            }
            // Remove the device.
            this.nvr.removeHomeKitDevice(protectDevice);
            return;
        }
    }
    // Listen to the UniFi Protect realtime updates API for updates we are interested in (e.g. motion).
    configureEvents() {
        // Only configure the event listener if it exists and it's not already configured.
        if (this.eventsHandler && this.ufpUpdatesHandler) {
            return true;
        }
        // Ensure we update our UFP state before we process any other events.
        this.prependListener("updateEvent", this.ufpUpdatesHandler = this.ufpUpdates.bind(this));
        // Process remove events.
        this.prependListener("addEvent", this.manageDevices.bind(this));
        // Listen for any messages coming in from our listener. We route events to the appropriate handlers based on the type of event that comes across.
        this.ufpApi.on("message", this.eventsHandler = (packet) => {
            switch (packet.header.action) {
                case "add":
                    this.emit("addEvent", packet);
                    if (packet.payload.camera) {
                        this.emit("addEvent." + packet.payload.camera, packet);
                    }
                    this.emit("addEvent." + packet.header.modelKey, packet);
                    break;
                case "remove":
                    this.emit("removeEvent", packet);
                    this.emit("removeEvent." + packet.header.id, packet);
                    this.emit("removeEvent." + packet.header.modelKey, packet);
                    break;
                case "update":
                    this.emit("updateEvent", packet);
                    this.emit("updateEvent." + packet.header.id, packet);
                    this.emit("updateEvent." + packet.header.modelKey, packet);
                    break;
                default:
                    break;
            }
            // If enabled, publish all the event traffic coming from the Protect controller to MQTT.
            if (this.mqttPublishTelemetry) {
                this.nvr.mqtt?.publish(this.nvr.ufp.mac, "telemetry", JSON.stringify(packet));
            }
        });
        return true;
    }
    // Motion event processing from UniFi Protect.
    motionEventHandler(protectDevice, lastMotion, detectedObjects = [], metadata) {
        if (!protectDevice || !lastMotion) {
            return;
        }
        // Only notify the user if we have a motion sensor and it's active.
        const motionService = protectDevice.accessory.getService(this.hap.Service.MotionSensor);
        if (motionService) {
            this.motionEventDelivery(protectDevice, motionService, detectedObjects, metadata);
        }
    }
    // Motion event delivery to HomeKit.
    motionEventDelivery(protectDevice, motionService, detectedObjects = [], metadata) {
        if (!protectDevice) {
            return;
        }
        // If we have disabled motion events, we're done here.
        if (("detectMotion" in protectDevice.accessory.context) && !protectDevice.accessory.context.detectMotion) {
            return;
        }
        // If we have an active motion event inflight, we're done.
        if (this.eventTimers[protectDevice.id]) {
            return;
        }
        const protectCamera = protectDevice;
        // Trigger the motion event in HomeKit.
        motionService.updateCharacteristic(this.hap.Characteristic.MotionDetected, true);
        // If we have a motion trigger switch configured, update it.
        protectDevice.accessory.getServiceById(this.hap.Service.Switch, ProtectReservedNames.SWITCH_MOTION_TRIGGER)?.updateCharacteristic(this.hap.Characteristic.On, true);
        // Publish the motion event to MQTT, if the user has configured it.
        this.nvr.mqtt?.publish(protectDevice.accessory, "motion", "true");
        // Log the event, if configured to do so.
        if (protectDevice.hints.logMotion) {
            let logMessage = "";
            // If we have license plate telemetry, let's inform the user if we're logging for motion events.
            if (metadata && ("licensePlate" in metadata) && detectedObjects.includes("licensePlate") && ("name" in metadata.licensePlate)) {
                logMessage = " (license plate: " + metadata.licensePlate.name + ", " + metadata.licensePlate.confidenceLevel + "% confidence)";
            }
            protectDevice.log.info("Motion detected%s.", ((protectDevice.ufp.modelKey === "camera") && detectedObjects.length &&
                (!protectCamera.stream?.hksv?.isRecording ||
                    protectCamera.hints.smartDetect || protectDevice.hasFeature("Motion.SmartDetect.ObjectSensors"))) ? ": " + detectedObjects.sort().join(", ") + logMessage : "");
        }
        // Iterate through our detected Protect smart object types.
        for (const detectedObject of detectedObjects) {
            // Trigger smart motion contact sensor, if configured.
            protectDevice.accessory.getServiceById(this.hap.Service.ContactSensor, ProtectReservedNames.CONTACT_MOTION_SMARTDETECT + "." + detectedObject)
                ?.updateCharacteristic(this.hap.Characteristic.ContactSensorState, true);
            // Publish the smart motion event to MQTT, if the user has configured it.
            this.nvr.mqtt?.publish(protectDevice.accessory, "motion/smart/" + detectedObject, "true");
            // Trigger license plate contact sensors, if configured.
            if (metadata && ("licensePlate" in metadata) && (detectedObject === "licensePlate") && ("name" in metadata.licensePlate)) {
                protectDevice.accessory.getServiceById(this.hap.Service.ContactSensor, ProtectReservedNames.CONTACT_MOTION_SMARTDETECT_LICENSE + "." +
                    metadata.licensePlate.name.toUpperCase())?.updateCharacteristic(this.hap.Characteristic.ContactSensorState, true);
                // Publish the smart motion event to MQTT, if the user has configured it.
                this.nvr.mqtt?.publish(protectDevice.accessory, "motion/smart/" + detectedObject + "/metadata", JSON.stringify(metadata));
            }
        }
        // Reset our motion event after motionDuration.
        this.eventTimers[protectDevice.id] = setTimeout(() => {
            motionService.updateCharacteristic(this.hap.Characteristic.MotionDetected, false);
            // If we have a motion trigger switch configured, update it.
            protectDevice.accessory.getServiceById(this.hap.Service.Switch, ProtectReservedNames.SWITCH_MOTION_TRIGGER)
                ?.updateCharacteristic(this.hap.Characteristic.On, false);
            protectDevice.log.debug("Resetting motion event.");
            // Publish to MQTT, if the user has configured it.
            this.nvr.mqtt?.publish(protectDevice.accessory, "motion", "false");
            // Delete the timer from our motion event tracker.
            delete this.eventTimers[protectDevice.id];
        }, protectDevice.hints.motionDuration * 1000);
        // If we have an active smart motion contact sensor reset timer, let's cancel it and create a new one.
        if (this.eventTimers[protectDevice.id + ".Motion.SmartDetect.ObjectSensors"]) {
            // Clear out the inflight timer from our motion event tracker.
            clearTimeout(this.eventTimers[protectDevice.id + ".Motion.SmartDetect.ObjectSensors"]);
        }
        // Reset our smart motion contact sensors after motionDuration.
        this.eventTimers[protectDevice.id + ".Motion.SmartDetect.ObjectSensors"] = setTimeout(() => {
            // Reset smart motion contact sensors, if configured.
            for (const detectedObject of detectedObjects) {
                // Reset our license plate contact sensor, if configured.
                if (metadata && ("licensePlate" in metadata) && (detectedObject === "licensePlate") && ("name" in metadata.licensePlate)) {
                    protectDevice.accessory.getServiceById(this.hap.Service.ContactSensor, ProtectReservedNames.CONTACT_MOTION_SMARTDETECT_LICENSE + "." +
                        metadata.licensePlate.name.toUpperCase())?.updateCharacteristic(this.hap.Characteristic.ContactSensorState, false);
                }
                // Reset our smart motion contact sensor, if configured.
                protectDevice.accessory.getServiceById(this.hap.Service.ContactSensor, ProtectReservedNames.CONTACT_MOTION_SMARTDETECT + "." + detectedObject)
                    ?.updateCharacteristic(this.hap.Characteristic.ContactSensorState, false);
                // Publish the smart motion event to MQTT, if the user has configured it.
                this.nvr.mqtt?.publish(protectDevice.accessory, "motion/smart/" + detectedObject, "false");
                protectDevice.log.debug("Resetting smart object motion event.");
            }
            // Delete the timer from our motion event tracker.
            delete this.eventTimers[protectDevice.id + ".Motion.SmartDetect.ObjectSensors"];
        }, protectDevice.hints.motionDuration * 1000);
        // If we don't have smart motion detection enabled, or if we do have it enabled and we have a smart motion event that's detected a person, let's process our
        // occupancy event updates.
        if (!protectDevice.hints.smartDetect || (protectDevice.hints.smartDetect && detectedObjects.some(x => protectDevice.hints.smartOccupancy.includes(x)))) {
            // First, let's determine if the user has an occupancy sensor configured, before we process anything.
            const occupancyService = protectDevice.accessory.getService(this.hap.Service.OccupancySensor);
            if (occupancyService) {
                // Kill any inflight reset timer.
                if (this.eventTimers[protectDevice.id + ".Motion.OccupancySensor"]) {
                    clearTimeout(this.eventTimers[protectDevice.id + ".Motion.OccupancySensor"]);
                }
                // If the occupancy sensor isn't already triggered, let's do so now.
                if (occupancyService.getCharacteristic(this.hap.Characteristic.OccupancyDetected).value !== true) {
                    // Trigger the occupancy event in HomeKit.
                    occupancyService.updateCharacteristic(this.hap.Characteristic.OccupancyDetected, true);
                    // Publish the occupancy event to MQTT, if the user has configured it.
                    this.nvr.mqtt?.publish(protectDevice.accessory, "occupancy", "true");
                    // Log the event, if configured to do so.
                    if (protectDevice.hints.logMotion) {
                        protectDevice.log.info("Occupancy detected%s.", protectDevice.hints.smartDetect ? ": " + protectDevice.hints.smartOccupancy.filter(x => detectedObjects.includes(x)).join(", ") : "");
                    }
                }
                // Reset our occupancy state after occupancyDuration.
                this.eventTimers[protectDevice.id + ".Motion.OccupancySensor"] = setTimeout(() => {
                    // Reset the occupancy sensor.
                    occupancyService.updateCharacteristic(this.hap.Characteristic.OccupancyDetected, false);
                    // Publish to MQTT, if the user has configured it.
                    this.nvr.mqtt?.publish(protectDevice.accessory, "occupancy", "false");
                    // Log the event, if configured to do so.
                    if (protectDevice.hints.logMotion) {
                        protectDevice.log.info("Occupancy no longer detected.");
                    }
                    // Delete the timer from our motion event tracker.
                    delete this.eventTimers[protectDevice.id];
                }, protectDevice.hints.occupancyDuration * 1000);
            }
        }
    }
    // Doorbell event processing from UniFi Protect and delivered to HomeKit.
    doorbellEventHandler(protectDevice, lastRing) {
        if (!protectDevice || !lastRing) {
            return;
        }
        // If we have an inflight ring event, and we're enforcing a ring duration, we're done.
        if (this.eventTimers[protectDevice.id + ".Doorbell.Ring"]) {
            return;
        }
        // Only notify the user if we have a doorbell.
        const doorbellService = protectDevice.accessory.getService(this.hap.Service.Doorbell);
        if (!doorbellService) {
            return;
        }
        // Trigger the doorbell. We delay this slightly to workaround what appears to be a race condition bug in HomeKit. Inelegant, but effective.
        setImmediate(() => {
            doorbellService.getCharacteristic(this.hap.Characteristic.ProgrammableSwitchEvent)
                ?.sendEventNotification(this.hap.Characteristic.ProgrammableSwitchEvent.SINGLE_PRESS);
        });
        // Check to see if we have a doorbell trigger switch configured. If we do, update it.
        const triggerService = protectDevice.accessory.getServiceById(this.hap.Service.Switch, ProtectReservedNames.SWITCH_DOORBELL_TRIGGER);
        if (triggerService) {
            // Kill any inflight trigger reset.
            if (this.eventTimers[protectDevice.id + ".Doorbell.Ring.Trigger"]) {
                clearTimeout(this.eventTimers[protectDevice.id + ".Doorbell.Ring.Trigger"]);
                delete this.eventTimers[protectDevice.id + ".Doorbell.Ring.Trigger"];
            }
            // Flag that we're ringing.
            protectDevice.isRinging = true;
            // Update the trigger switch state.
            triggerService.updateCharacteristic(this.hap.Characteristic.On, true);
            // Reset our doorbell trigger.
            this.eventTimers[protectDevice.id + ".Doorbell.Ring.Trigger"] = setTimeout(() => {
                protectDevice.isRinging = false;
                triggerService.updateCharacteristic(this.hap.Characteristic.On, false);
                this.log.debug("Resetting doorbell ring trigger.");
                // Delete the timer from our motion event tracker.
                delete this.eventTimers[protectDevice.id + ".Doorbell.Ring.Trigger"];
            }, PROTECT_DOORBELL_TRIGGER_DURATION);
        }
        // Publish to MQTT, if the user has configured it.
        this.nvr.mqtt?.publish(protectDevice.accessory, "doorbell", "true");
        if (protectDevice.hints.logDoorbell) {
            protectDevice.log.info("Doorbell ring detected.");
        }
        // Kill any inflight MQTT reset.
        if (this.eventTimers[protectDevice.id + ".Doorbell.Ring.MQTT"]) {
            clearTimeout(this.eventTimers[protectDevice.id + ".Doorbell.Ring.MQTT"]);
            delete this.eventTimers[protectDevice.id + ".Doorbell.Ring.MQTT"];
        }
        // Fire off our MQTT doorbell ring event.
        this.eventTimers[protectDevice.id + ".Doorbell.Ring.MQTT"] = setTimeout(() => {
            this.nvr.mqtt?.publish(protectDevice.accessory, "doorbell", "false");
            // Delete the timer from our event tracker.
            delete this.eventTimers[protectDevice.id + ".Doorbell.Ring.MQTT"];
        }, PROTECT_DOORBELL_TRIGGER_DURATION);
        // If we don't have a ring duration defined, we're done.
        if (!this.nvr.platform.config.ringDelay) {
            return;
        }
        // Reset our ring threshold.
        this.eventTimers[protectDevice.id + ".Doorbell.Ring"] = setTimeout(() => {
            // Delete the timer from our event tracker.
            delete this.eventTimers[protectDevice.id + ".Doorbell.Ring"];
        }, this.nvr.platform.config.ringDelay * 1000);
    }
}
//# sourceMappingURL=protect-nvr-events.js.map